%% Ein einfaches inverses Problem: Masseverteilung und Gravitationsfeld
% Umkehrproblem der Potentialtheorie, vereinfacht in 2 Dimensionen.
% 
% Hintergrund-Info (Quelle: <https://de.wikipedia.org/wiki/Umkehrproblem_der_Potentialtheorie 
% https://de.wikipedia.org/wiki/Umkehrproblem_der_Potentialtheorie>) 
% 
% Das Umkehrproblem der Potentialtheorie ist ein theoretisch begründetes Mehrdeutigkeitsproblem 
% bei einigen Anwendungen der Potentialtheorie. Es begrenzt vor allem die Interpretation 
% von Messdaten in einem Magnet- oder Schwerefeld, weil es einen eindeutigen Rückschluss 
% vom Kraftfeld auf die verursachenden Quellpunkte verhindert. Das Umkehrproblem 
% erhielt seinen Namen in Analogie zur Umkehr von Ursache und Wirkung. Bei einer 
% bekannten oder vorgegebenen Massenverteilung im Innern eines Festkörpers lässt 
% sich deren Wirkung nach außen streng berechnen, während umgekehrt das gemessene 
% Schwerefeld durch verschiedene Massenanordnungen hervorgerufen sein kann.

clear all
%% Zuerst das direkte Problem (auch: Vorwärts-Problem)
% Gegeben ist die Masseverteilung im Untergrund. Gesucht ist das resultierende 
% Gravitationsfeld an der Oberfläche.
% 
% Annahme: Ebene Oberfläche, Koordinate x entlang der Oberfläche, Koordinate 
% z zeigt in die Tiefe. Eine homogene Masseverteilung im Untergrund erzeugt ein 
% homogenes Gravitationsfeld an der Oberfläche. Abweichende Masseverteilung, zum 
% Beispiel ein schwerer Störkörper, bewirkt Störungen im Gravitationsfeld. Wir 
% modellieren hier diese Störungen. Wir rechnen in passend skalierten Einheiten, 
% so dass Gravitationskonstante = 1 gilt.

% Das Gebiet, in dem sich die Störmassen verteilen, wird mit einem
% Rechengitter überzogen. Die Störmassen werden als Punktmassen an der
% Gitterpunkten modelliert und deren Beiträge aufsummiert. Das ist eine
% Näherung für das Integral bei kontinuierlicher Massenverteilung.

% Auflösung: Anzahl Gitterpunkte in x- und z-Richtung
nx = 6; 
nz = 5;   
% Koordinaten der Gitterpunkte
x = linspace(0,4,nx);
z =  linspace(0,1,nz) + 1/(nz-1)/2;
[X,Z] = meshgrid(x,z);
%% hier werden Störmassen in einigen Teilgebieten definiert

M = zeros(size(X));
s1=find(abs(X-3)<=0.5 & abs(Z-0.8)<=0.3);
M(s1)=1;

s2=find((X-1 ).^2 + (Z-0.8).^2 <=0.5^2);
M(s2)=-0.5;

figure(1)
imagesc(x,z,M)
title('Störmasseverteilung im Untergrund')
ylabel('Tiefe z')
xlabel('Horizontaldistanz x')
colorbar
%% Gravitation entlang x-Achse
% kann im Vorwärts-Problem in feiner Auflösung berechnet werden

xx = linspace(-3,8,200);
gg = zeros(size(xx));
for i=1:length(xx)
    % Beitrag der Störmassen im Punkt x=xi
    U = M./((xx(i) - X).^2 + Z.^2);
    gg(i)=sum(sum(U));
end

figure(2)
plot(xx,gg)
title('Gravitations-Störung entlang Oberfläche')
ylabel('Tiefe z')
xlabel('Horizontaldistanz x')
%% Das inverse Problem
% Gegeben sind nun Messungen der Gravitation an der Oberfläche, gesucht ist 
% die Masseverteilung im Untergrund. Hier verwenden wir die soeben berechneten 
% gg-Daten, um die Massenverteilung an den x-z-Gitterpunkten zu bestimmen.

% Unbekannt sind die nx*nz Massenwerte an den Gitterpunkten. Es liegen dafür 
% length(xx) Messwerte vor. Matrix A des resultierenden Gleichungssystems
% hat length(xx) Zeilen und nx*nz Spalten.
A = zeros(length(xx),nx*nz);

for i=1:length(xx)
    % Gleichung für den Datenpunkt xx(i). Die Koeffizienten werden zuerst
    % als Matrix bestimmt, aber sie werden dann als Zeilenvektor
    % zusammengefasst und in die Matrix des Gleichungssystems eingetragen.
    U = 1./((xx(i) - X).^2 + Z.^2);
    A(i,:)=U(:);
end

% Lösung des überbestimmten Systems: bei genügend viel exakten Daten lässt sich die
% Masseverteilung unter Umständen rekonstruieren, aber nur, wenn A vollen
% Spaltenrang hat. Bei grösseren Problemen ist das nicht der Fall; dann
% versagt dies Rekonstruierung, weil es keine eindeutige Lösung gibt.
m = A\gg';

% Unterschied Original - rekonstruierte Verteilung: ist in diesem sehr
% einfachen Beispiel vernachlässigbar klein, wenn die Daten völlig exakt
% vorliegen. Aber...
norm(m-M(:))

% Messwerte sind nie exakt. Kleine Störungen sind unvermeidbar. Hier werden
% die Originalwerte leicht gestört. Ohne Regularisierung ruiniert das die
% Rekonstruktion.
gg_e = gg + 0.0001*norm(gg,'inf')*randn(size(gg));

regularisierung = 'none';
%regularisierung = 'Tichonow SVD';
%regularisierung = 'Tichonow NormGl';

M_e=zeros(size(M));
alpha=0.0001;
switch regularisierung
    case 'none'
        % Kleinste-Quadrate-Ausgleichslösung
        m_e = A\gg_e';
        M_e(:)=m_e;
    case 'Tichonow SVD'
        [U, S, V]=svd(A);
        Sreg= diag(S)./(diag(S).^2 + alpha);
        tmp=(U'*gg_e');
        m_regSVD = V*(Sreg.*tmp(1:nx*nz));
        M_e(:)=m_regSVD;
    case 'Tichonow NormGl'
        % ist aequivalent zur SVD-Methode; Ältere Form. Heisst auch ridge
        % regression.
        AtA = A'*A + alpha*eye(nx*nz);
        m_regNGL =  AtA\A'*gg_e';
        M_e(:)=m_regNGL;
end


figure(3)
imagesc(x,z,M_e)
title('Rekonstruierte Störmasseverteilung im Untergrund')
ylabel('Tiefe z')
xlabel('Horizontaldistanz x')
colorbar