%% Zweidimensionale FFT mixt Bilder
% Dieses Beispiel illustriert 2D FFT angewandt auf Fotos. Zum Spass vertauscht 
% es Amplitude und Phasen-Information der beiden Bilder. Die Absicht ist, spielerisch 
% etwas Verständnis für FFT und die Information, die in Frequenz- und Amplitutenanteil 
% steckt, zu vermitteln.
% 
% Verwendete MATLAB-Befehle im Zusammenhang mit FFT: fft2, ifft2, fftshift, 
% abs, angle
% 
% Außedrem lernen Sie hier einige MATLAB-Befehle im Zusammenhang mit Bildbearbeitung 
% kennen: imread, imshow, im2gray, imagesc, Datentyp uint8
% 
% C.Brand, E. Hausenblas 2022. Ideen zu dieser Datei stammen von:
% 
% <https://matlabgeeks.com/tips-tutorials/how-to-do-a-2-d-fourier-transform-in-matlab/ 
% https://matlabgeeks.com/tips-tutorials/how-to-do-a-2-d-fourier-transform-in-matlab/>
% 
% A Guide to MATLAB: For Beginners and Experienced Users von Brian R. Hunt, 
% Ronald L. Lipsman, Jonathan M. Rosenberg, Kapitel Image compression and manipulation. 
% <https://books.google.at/books?id=nwAPBAAAQBAJ&pg=PA175&lpg=PA175#v=onepage&q&f=false 
% https://books.google.at/books?id=nwAPBAAAQBAJ&pg=PA175&lpg=PA175#v=onepage&q&f=false> 
%% Bilder einlesen und anzeigen
% Die beiden Bilder müssen gleiche Pixel-Dimensionen haben!

clear; close all
bildA = imread('wueste1.jpg'); % (Quelle: E. Hausenblas)
bildB = imread('blumen.jpg');

figure;imshow(bildA)
title('Bild A - Wüste')

figure;imshow(bildB)
title('Bild B - Blumen')
%% Zuerst mal für die Schwarz-Weiss-Versionen
% Der Einfachheit halber arbeiten wir zuerst einmal mit Schwarzweiss-Versionen 
% der beiden Bilder.

swBildA = im2gray(bildA);
% Und wir schneiden Teilbereiche aus den Bildern aus.
swBildA = swBildA(1:1000,end-999:end);
figure; imshow(swBildA)

swBildB = im2gray(bildB);
swBildB = swBildB(end-999:end,1501:2500);
figure; imshow(swBildB)
%% Die 2D-Fouriertransformation der Bilder
% fft2 besorgt die 2D-Fouriertransformation. Das Ergebnis ist eine Matrix komplexer 
% Zahlen. Wir lassen Betrag (Amplitude) und Phasenwinkel anzeigen. fftshift ordnet 
% das Ergebnis von fft2 noch um, so dass die Information zu den niedrigen Frequenzen 
% im Zentrum der Bilder sichtbar werden. Außerdem stellen wir die Amplituden in 
% logarithmischem Maßstab dar, weil diese über mehrere Größenordnungen variieren.

fftA = fft2(swBildA);
figure; imagesc(log(abs(fftshift(fftA)))); 
colormap gray;axis equal tight
title('Fourier-Amplitude - Wüste')
% Weil fftA double-Werte enthält, brauchen wir *|imagesc|* statt *|imshow|*
% zum Anzeigen.

fftB = fft2(swBildB);
figure; imagesc(log(abs(fftshift(fftB)))); 
colormap gray;axis equal tight
title('Fourier-Amplitude - Blumen')

% Interpretation: Amplitude (genauer: Quadrat der Amplitude, aber im
% logarithmischen Maßstab ist das bis auf einen multiplikativen
% Skalierungsfaktor dasselbe) entspricht Energie oder Informationsgehalt.
% Man erkennt: Der Großteil der Bildinformation konzentriert sich in der
% Mitte und entlang des Achsenkreuzes, das heißt, bei den niedrigen
% Frequenzen. Das entspricht den eher großflächigen Bildstrukturen. Feine,
% kontrastreiche Details im Bild erzeugen hochfrequente Beiträge, die
% weiter weg vom Zentrum angezeigt werden.
%
% Scharfe Kanten im Bild erzeugen in der FFT Strahlen, die im rechten
% Winkel zu den Kanten im Bild orientiert sind. Beim Wüstenbild erzeugen
% Horizont und die schräg abwärts verlaufenden Schattenmuster Strukturen in
% der FFT. Beim Blumen-Bild sieht die FFT symmetrischer aus, weil hier
% Kanten in unterschiedlichen Richtungen eher gleichmäßig verteilt und eher
% unscharf sind.

% Die Phasenwinkel der FFTs sind in der bildlichen Darstellung nicht
% wirklich interpretierbar. Sie sehen zwar wie unregelmäßiges Rauschen aus,
% sind aber doch für die beiden Bilder deutlich unterschiedlich. (Die
% Phasenwinkel variieren, unabhängig ob bei niedrigen oder hohen
% Frequenzen, zwischen +-pi, daher ist hier keine logarithmische
% Darstellung notwendig.)

figure; imagesc(angle(fftshift(fftA))); 
colormap gray;axis equal tight
title('Fourier-Phase - Wüste')

figure; imagesc(angle(fftshift(fftB))); 
colormap gray;axis equal tight
title('Fourier-Phase - Blumen')
%% Rücktransformation mit vertauschter Phasen-Information
% Die Phase enthält einen ganz wesentlichen Teil der Information. Um das zu 
% demonstrieren, werden hier Amplitude und Phase der beiden Bilder vertauscht 
% und danach mit inverser FFT rücktransformiert. Die zwei Mischbilder werden angezeigt.

fftC = abs(fftA).*exp(1i*angle(fftB));
fftD = abs(fftB).*exp(1i*angle(fftA));
 

bildC = ifft2(fftC);
BildD = ifft2(fftD);
figure; imagesc(abs(bildC)); 
colormap gray;axis equal tight
title('Mix: Amplitude Wüste; Phase Blumen')

figure; imagesc(abs(BildD)); 
colormap gray;axis equal tight
title('Mix: Amplitude Blumen; Phase Wüste')
%% Und jetzt nochmal in der Farb-Version
% und auf die gesamten Bilder angewandt. Dazu muss die FFT, der Phasenmix und 
% die Rücktransformation einzeln auf die drei Farbkanäle angewandt werden.

fftAred = fft2(bildA(:,:,1));
fftAgrn = fft2(bildA(:,:,2));
fftAblu = fft2(bildA(:,:,3));

fftBred = fft2(bildB(:,:,1));
fftBgrn = fft2(bildB(:,:,2));
fftBblu = fft2(bildB(:,:,3));

fftCred = abs(fftAred).*exp(1i*angle(fftBred));
fftCgrn = abs(fftAgrn).*exp(1i*angle(fftBgrn));
fftCblu = abs(fftAblu).*exp(1i*angle(fftBblu));
 
bildC=zeros(size(bildA),'uint8');

% Bilddaten enthalten unsigned 8-bit integers, also Werte 0-255. Die
% Rücktrtansformation der gemixten Fourier-Werte sollte theoretisch 
% reell sein; aufgrund von Rundungsfehlern entsteht aber Datentyp komplex
% mit Imaginärteilen bis auf Rundungsfehler gleich 0. abs macht daraus rein
% reelle double-Werte. In den Original-Bildern ist der Bereich 0-255. Bei den
% gemixten Bildern stimmt das nicht ganz. uint8 konvertiert auf den
% richtigen Datentyp und setzt dabei Werte > 255 auf 255.

bildC(:,:,1) = uint8(abs(ifft2(fftCred)));
bildC(:,:,2) = uint8(abs(ifft2(fftCgrn)));
bildC(:,:,3) = uint8(abs(ifft2(fftCblu)));
figure; imshow(bildC); 
title('Mix: Amplitude Wüste; Phase Blumen')

fftDred = abs(fftBred).*exp(1i*angle(fftAred));
fftDgrn = abs(fftBgrn).*exp(1i*angle(fftAgrn));
fftDblu = abs(fftBblu).*exp(1i*angle(fftAblu));
 
bildD=zeros(size(bildA),'uint8');
bildD(:,:,1) = uint8(abs(ifft2(fftDred)));
bildD(:,:,2) = uint8(abs(ifft2(fftDgrn)));
bildD(:,:,3) = uint8(abs(ifft2(fftDblu)));
figure; imshow(bildD); 
title('Mix: Amplitude Blumen; Phase Wüste')